#include <array>

#include <error.hh>
#include <signal.hh>

#include "macros.h"

namespace fx::signal {
	namespace {
		template<typename T>
			requires(std::is_trivial_v<T>)
		constexpr static bool biteq(T a, T b) noexcept
		{
			typedef std::array<unsigned char, sizeof(T)>  A alignas(T);
			__builtin_clear_padding(&a);
			__builtin_clear_padding(&b);

			return std::bit_cast<A>(a) == std::bit_cast<A>(b);
		}
	}

#undef IMPL_SPEC
#define IMPL_CLASS wait_for_signals
#define IMPL_SPEC <void>
	$ctor(std::initializer_list<int> sigs, with_add_t const&, std::initializer_list<int> add) noexcept
	{
		sigset_t cur_set = {0};
		sigset_t add_set = {0};

		sigemptyset(&cur_set);
		for(const int sig : sigs)
			sigaddset(&cur_set, sig);

		sigemptyset(&add_set);
		for(const int sig : add)
			sigaddset(&add_set, sig);

		sigemptyset(&sigset);
		sigorset(&sigset, &cur_set, &add_set);


		pthread_sigmask(SIG_BLOCK, &sigset, &oldset);
	}

	int $ wait_for_signal(sigset_t const& sigset)
	{
		int signum = 0;
		if(__builtin_expect(sigwait(&sigset, &signum) != 0, false))
			fx::error::throw_last_system_error();
		return signum;
	}

	void $ do_signal_reached(int signum) &  { (void)signum; }
	void $ do_signal_reached(int signum) && { (void)signum; }

	bool $ do_reset_mask() noexcept 
	{
		// If the oldset is not empty (i.e. object has not been moved,) then we re-set the mask to the oldset IFF `sigset` is also not empty. If it is, we do not know if the object *was* move, but *nothing will have changed* in the mask, so no need to re-set the mask anyway.
		if((not sigisemptyset(&oldset)) || __builtin_expect(! biteq(oldset, sigset), true))
			// XXX: I'm not sure if this ^ check is actually what we want wrt. moving, but for non-moved handles, it seems to work fine. (See thread-spawning version for issues.)
			pthread_sigmask(SIG_SETMASK, &oldset, &sigset);
		else return false;
		return true;
	}

	int $ operator()(sigset_t const& sigset) &
	{
		int signum = wait_for_signal(sigset);
		
		do_signal_reached(signum);
		return signum;
	}
	int $ operator()(sigset_t const& sigset) &&
	{
		int signum = wait_for_signal(sigset);

		std::move(*this).do_signal_reached(signum);
		return signum;
	}


	$ctor(std::initializer_list<int> sigs) noexcept
	{	
		sigemptyset(&sigset);
		for(const int sig : sigs)
			sigaddset(&sigset, sig);
		pthread_sigmask(SIG_BLOCK, &sigset, &oldset);
	}

	$ctor_move(m) noexcept
		: oldset(m.oldset)
		, sigset(m.sigset) {
			// Ensure `m.oldset` is emptied so dtor doesn't try to unblock it and knows it's been moved.
			sigemptyset(&m.oldset);
		}

	/// Do not reset the signal mask on destruction.
	///
	/// This will leave the blocked signals handled by the invocation operator blocked and `sigwait()` can still be used on them.
	void $ detach() && noexcept {
		sigemptyset(&oldset);
	}

	/// Reset the signal mask right now.
	///
	/// This also detaches the handle from doing it on destruction (see `detach()`.)
	bool $ reset_mask() noexcept {
		bool v = do_reset_mask();
		std::move(*this).detach();
		return v;
	}

	$dtor() noexcept
	{
		do_reset_mask();
	}

	int $ operator()() &
       	{ return this->operator()(this->sigset); }
	int $ operator()() && 
	{ return std::move(*this)(this->sigset); }

#undef IMPL_SPEC  // template<> <void>

	template class $class<void>;
	template class $class<void(*)(int)>;
	template class $class< std::function< void(int) > >;

#undef IMPL_CLASS // wait_for_signals
#define IMPL_SPEC // 
	
#define IMPL_CLASS thread_wait_for_signals
	extern template class $class< void(*)(int) >;
	extern template class $class< void(*)(int, std::stop_token) >;

	extern template class $class<std::function<void(int)> >;
	extern template class $class<std::move_only_function<void(int)> >;

	extern template class $class<std::function<void(int, std::stop_token)> >;
	extern template class $class<std::move_only_function<void(int, std::stop_token)> >;
#undef IMPL_CLASS /* thread_wait_for_signals */

}

